from typing import Dict

from google.protobuf.json_format import MessageToDict
from yupy import (
    mapping,
    array,
    string,
    number,
    mixed,
    union,
    ValidationError,
    Constraint,
    required,
)

from a7p import profedit_pb2

__all__ = ("validate",)

_schema = mapping().shape(
    {
        "profile": required(
            mapping().shape(
                {
                    # descriptor
                    "profile_name": required(
                        string().max(50), "Profile name is required"
                    ),
                    "cartridge_name": required(
                        string().max(50), "Cartridge name is required"
                    ),
                    "bullet_name": required(
                        string().max(50), "Bullet name is required"
                    ),
                    "short_name_top": required(
                        string().max(8), "Short name top is required"
                    ),
                    "short_name_bot": required(
                        string().max(8), "Short name bottom is required"
                    ),
                    "caliber": required(string().max(50), "Caliber is required"),
                    "device_uuid": string().max(50),
                    "user_note": string().max(1024),
                    # zeroing
                    "zero_x": required(
                        number().ge(-200000).le(200000).integer(), "Zero X is required"
                    ),
                    "zero_y": required(
                        number().ge(-200000).le(200000).integer(), "Zero Y is required"
                    ),
                    # lists
                    "distances": required(
                        array()
                        .of(number().ge(100).le(300000).integer())
                        .min(1)
                        .max(200)
                    ),
                    "switches": array()
                    .of(
                        union().one_of(
                            (
                                mapping().shape(
                                    {
                                        "c_idx": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                        "distance_from": required(
                                            mixed().one_of(["VALUE"])
                                        ),
                                        "distance": required(
                                            number().ge(100).le(300000).integer()
                                        ),
                                        "reticle_idx": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                        "zoom": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                    }
                                ),
                                mapping().shape(
                                    {
                                        "c_idx": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                        "distance_from": required(
                                            mixed().one_of(["INDEX"])
                                        ),
                                        "distance": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                        "reticle_idx": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                        "zoom": required(
                                            number().ge(0).le(255).integer()
                                        ),
                                    }
                                ),
                            )
                        )
                    )
                    .min(4),
                    # rifle
                    "sc_height": required(number().ge(-5000).le(5000).integer()),
                    "r_twist": required(number().ge(0).le(10000).integer()),
                    "twist_dir": required(mixed().one_of(["RIGHT", "LEFT"])),
                    # cartridge
                    "c_muzzle_velocity": required(number().ge(100).le(30000).integer()),
                    "c_zero_temperature": required(number().ge(-100).le(100).integer()),
                    "c_t_coeff": required(number().ge(0).le(5000).integer()),
                    # zero params
                    "c_zero_distance_idx": required(number().ge(0).le(255).integer()),
                    "c_zero_air_temperature": required(
                        number().ge(-100).le(100).integer()
                    ),
                    "c_zero_air_pressure": required(
                        number().ge(3000).le(15000).integer()
                    ),
                    "c_zero_air_humidity": required(number().ge(0).le(100).integer()),
                    "c_zero_w_pitch": required(number().ge(-90).le(90).integer()),
                    "c_zero_p_temperature": required(
                        number().ge(-100).le(100).integer()
                    ),
                    # bullet
                    "b_diameter": required(number().ge(1).le(50000).integer()),
                    "b_weight": required(number().ge(10).le(65535).integer()),
                    "b_length": required(number().ge(1).le(200000).integer()),
                    # drag model
                    "bc_type": required(mixed().one_of(["G1", "G7", "CUSTOM"])),
                    "coef_rows": required(array().min(1).max(200)),
                }
            )
        )
    }
)


def _coef_rows_mv_validator(rows):
    if not rows:
        raise ValidationError(
            Constraint("unique_mv", "Coefficient rows must not be empty"),
            invalid_value=rows,
        )
    mv_values = map(lambda x: x["mv"], rows)
    filtered = tuple(filter(lambda mv: mv != 0, mv_values))
    unique = set(filtered)
    if len(unique) != len(filtered):
        raise ValidationError(
            Constraint("unique_mv", "'mv' values must be unique, except for mv == 0"),
            invalid_value=rows,
        )


_coef_rows_std_schema = required(
    array()
    .of(
        mapping().shape(
            {
                "bc_cd": number().ge(0).le(10000).integer(),
                "mv": number().ge(0).le(30000).integer(),
            }
        )
    )
    .min(1)
    .max(5)
    .test(_coef_rows_mv_validator),
    "For G1 or G7, coefRows must contain between 1 and 5 items",
)

_coef_rows_custom_schema = required(
    array()
    .of(
        mapping().shape(
            {
                "bc_cd": number().ge(0).le(10000).integer(),
                "mv": number().ge(0).le(10000).integer(),
            }
        )
    )
    .min(1)
    .max(200)
    .test(_coef_rows_mv_validator),
    "For CUSTOM, coefRows must contain between 1 and 200 items",
)


def validate_coef_rows(data: Dict, fail_fast):
    bc_type = data["profile"]["bc_type"]
    match bc_type:
        case "G1" | "G7":
            _coef_rows_std_schema.validate(
                data["profile"]["coef_rows"], fail_fast, path="~/profile/coef_rows"
            )
        case "CUSTOM":
            _coef_rows_custom_schema.validate(
                data["profile"]["coef_rows"], fail_fast, path="~/profile/coef_rows"
            )


def validate(payload: profedit_pb2.Payload, fail_fast: bool = False):
    data = MessageToDict(
        payload, including_default_value_fields=True, preserving_proto_field_name=True
    )
    errors = []
    try:
        _schema.validate(data, fail_fast, "~")
    except ValidationError as err:
        if fail_fast:
            raise err
        errors.append(err)

    try:
        validate_coef_rows(data, fail_fast)
    except ValidationError as err:
        if fail_fast:
            raise err
        errors.append(err)

    if errors:
        raise ValidationError(
            Constraint("invalid payload", "invalid payload"),
            path="~",
            errors=errors,
            invalid_value=payload,
        )



"""
Stubs for profedit_pb2
"""

from typing import List, Union
from google.protobuf.message import Message


class Payload(Message):
    profile: 'Profile'


class CoefRow(Message):
    bc_cd: int
    mv: int


DType = Union[int, str]


class SwPos(Message):
    c_idx: int
    reticle_idx: int
    zoom: int
    distance: int
    distance_from: DType


class Profile(Message):
    profile_name: str
    cartridge_name: str
    bullet_name: str
    short_name_top: str
    short_name_bot: str
    user_note: str
    zero_x: int
    zero_y: int
    sc_height: int
    r_twist: int
    c_muzzle_velocity: int
    c_zero_temperature: int
    c_t_coeff: int
    c_zero_air_temperature: int
    c_zero_air_pressure: int
    c_zero_air_humidity: int
    c_zero_p_temperature: int
    c_zero_w_pitch: int
    b_diameter: int
    b_weight: int
    b_length: int
    twist_dir: 'TwistDir'
    bc_type: 'GType'
    switches: List['SwPos']
    distances: List[int]
    c_zero_distance_idx: int
    coef_rows: List['CoefRow']
    caliber: str


class TwistDir:
    RIGHT = 0
    LEFT = 1


class GType:
    G1 = 0
    G7 = 1
    CUSTOM = 2



import argparse
import logging
from pathlib import Path

import a7p
from a7p import pydantic
from a7p.logger import color_print

parser = argparse.ArgumentParser(prog="a7p-recover")

parser.add_argument("path", help="Path to broken .a7p file", type=Path)

logger = logging.getLogger("a7p")
logger.setLevel(logging.DEBUG)


def main():
    args = parser.parse_args()
    file_path: Path = args.path
    if not file_path.exists():
        parser.error(f"No such file or directory: '{file_path}'")

    with open(file_path.absolute(), "rb") as fp:
        payload = a7p.load(fp, validate_=False)

        model, restored, violations = pydantic.validate(payload)
        for v in violations:
            color_print(v.format(), levelname="WARNING")

        if violations:
            logger.info("Started restore process")
            model, restored, violations = pydantic.validate(payload, restore=True)
            for r in restored:
                r.print()

            if violations:
                for v in violations:
                    color_print(v.format(), levelname="WARNING")
            else:
                dump = model.model_dump(mode="json")

                with open(
                    file_path.with_stem(file_path.stem + "_restored"), "wb"
                ) as fp:
                    restored_payload = a7p.from_dict(dump)
                    a7p.dump(restored_payload, fp)


if __name__ == "__main__":
    main()



from enum import Enum
from typing import Any, get_args

from pydantic import (
    BaseModel,
    ValidationError,
    conint,
    constr,
    conlist,
    BeforeValidator,
    field_validator,
)
from pydantic_core.core_schema import FieldValidationInfo
from typing_extensions import List, Union, Annotated

from a7p.pydantic.correction import (
    on_restore,
    trigger_confield_validation,
    pre_validate_conint,
)
from a7p.pydantic.template import PAYLOAD_RECOVERY_SCHEMA


def validate_c_idx(v):
    if not (0 <= v <= 200 or v == 255):
        raise ValueError("c_idx must be between 0 and 200, or exactly 255")
    # assert 0 <= v <= 200 or v == 255
    return v


def validate_distance_from(v):
    if isinstance(v, int):
        # Ensure the integer is within the allowed range
        if not (0 <= v <= 255):
            raise ValueError(
                "distance_from must be between 0 and 255 if it's an integer."
            )
    elif isinstance(v, str):
        # Ensure the string is 'VALUE'
        if v.lower() not in ["value", "index"]:
            raise ValueError("distance_from must be 'VALUE' if it's a string.")
    else:
        raise ValueError(
            "distance_from must be either an integer in range 0-255 or the string 'VALUE' or 'INDEX'."
        )

    return v


def restore_str_len(max_len, default="nil"):
    def decorator(
        cls: type, value: Any, info: FieldValidationInfo, err: Exception
    ) -> Any:
        if info.context.get("restore"):
            if isinstance(value, str):
                return value[:max_len]
        return default[:max_len]

    return decorator


def restore_default(default):
    def decorator(
        cls: type, value: Any, info: FieldValidationInfo, err: Exception
    ) -> Any:
        return default

    return decorator


class BCType(Enum):
    G1 = "G1"
    G7 = "G7"
    CUSTOM = "CUSTOM"


class TwistDir(Enum):
    RIGHT = "RIGHT"
    LEFT = "LEFT"


class Switch(BaseModel):
    c_idx: Annotated[int, BeforeValidator(validate_c_idx)]
    zoom: conint(ge=0, le=6)
    distance: conint(ge=int(1.0 * 100), le=int(3000.0 * 100))
    reticle_idx: conint(ge=0, le=255)
    distance_from: Annotated[Union[int, str], BeforeValidator(validate_distance_from)]


class CoefRows(BaseModel):
    bc_cd: int
    mv: int


class CdMaRows(CoefRows):
    bc_cd: conint(ge=int(0.0 * 10000), le=int(10.0 * 10000))
    mv: conint(ge=int(0.0 * 10000), le=int(10.0 * 10000))


class BcMvRows(CoefRows):
    bc_cd: conint(ge=int(0.0 * 10000), le=int(10.0 * 10000))
    mv: conint(ge=int(0.0 * 10), le=int(3000.0 * 10))


class ProfileValidationError(ValueError):
    def __init__(self, message: str, errors: str):
        super().__init__(message)
        self.errors = errors


LongString = constr(max_length=50)
ShortString = constr(max_length=8)
Text = constr(max_length=1024)

SightHeight = conint(ge=-5000, le=5000)
Twist = conint(ge=0, le=100 * 100)

Velocity = conint(ge=-10 * 10, le=3000 * 10)
TCoeff = conint(ge=0, le=5 * 1000)

Zeroing = conint(ge=-200 * 1000, le=200 * 1000)
Temperature = conint(ge=-100, le=100)
Pressure = conint(ge=300 * 10, le=1500 * 10)
Humidity = conint(ge=0, le=100)
Pitch = conint(ge=-90 * 10, le=90 * 10)

Diameter = conint(ge=int(0.001 * 1000), le=int(50.0 * 1000))
Weight = conint(ge=int(1.0 * 10), le=int(6553.5 * 10))
Length = conint(ge=int(0.01 * 1000), le=int(200.0 * 1000))

SwitchesList = conlist(Switch, min_length=4)
Distance = conint(ge=int(1.0 * 100), le=int(3000.0 * 100))
DistancesList = conlist(Distance, min_length=1, max_length=200)

DistanceIdx = conint(ge=0, le=200, strict=True)
ZeroDistanceIdx = DistanceIdx
# CoefRowsList = Annotated[Union[List[BcMvRows], List[CdMaRows]], BeforeValidator(validate_coef_rows_based_on_bc_type)]
CoefRowsList = Union[List[BcMvRows], List[CdMaRows]]

DEFAULT_DISTANCES = PAYLOAD_RECOVERY_SCHEMA["profile"]["distances"]
DEFAULT_SWITCHES = PAYLOAD_RECOVERY_SCHEMA["profile"]["switches"]
DEFAULT_COEF_ROWS = PAYLOAD_RECOVERY_SCHEMA["profile"]["coef_rows"]


class Profile(BaseModel):
    # description params
    profile_name: LongString
    cartridge_name: LongString
    bullet_name: LongString
    short_name_top: ShortString
    short_name_bot: ShortString
    caliber: LongString
    device_uuid: LongString
    user_note: Text

    # zeroing props
    zero_x: Zeroing
    zero_y: Zeroing

    # barrel params
    sc_height: SightHeight
    r_twist: Twist
    twist_dir: TwistDir

    # muzzle velocity
    c_muzzle_velocity: Velocity
    c_t_coeff: TCoeff

    # zero environment props
    c_zero_temperature: Temperature
    c_zero_air_temperature: Temperature
    c_zero_p_temperature: Temperature
    c_zero_air_pressure: Pressure
    c_zero_air_humidity: Humidity
    c_zero_w_pitch: Pitch

    # bullet params
    b_diameter: Diameter
    b_weight: Weight
    b_length: Length

    switches: SwitchesList
    distances: DistancesList
    c_zero_distance_idx: ZeroDistanceIdx
    bc_type: BCType
    coef_rows: CoefRowsList

    @field_validator(
        "profile_name", "cartridge_name", "bullet_name", "caliber", mode="before"
    )
    @on_restore(handler=restore_str_len(50))
    def validate_string_50(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("device_uuid", mode="before")
    @on_restore(handler=restore_str_len(50, ""))
    def validate_device_uuid(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("short_name_top", "short_name_bot", mode="before")
    @on_restore(handler=restore_str_len(8))
    def validate_string_8(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("user_note", mode="before")
    @on_restore(handler=restore_str_len(1024))
    def validate_user_note(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator(
        "zero_x", "zero_y", "c_zero_air_humidity", "c_zero_w_pitch", mode="before"
    )
    @on_restore(handler=restore_default(0))
    def validate_zero(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("sc_height", mode="before")
    @on_restore(handler=restore_default(90))
    def validate_sc_height(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("c_muzzle_velocity", mode="before")
    @on_restore(handler=restore_default(8000))
    def validate_c_muzzle_velocity(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator(
        "c_zero_temperature",
        "c_zero_air_temperature",
        "c_zero_p_temperature",
        mode="before",
    )
    @on_restore(handler=restore_default(15))
    def validate_temperature(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("c_t_coeff", mode="before")
    @on_restore(handler=restore_default(1000))
    def validate_c_t_coeff(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("c_zero_air_pressure", mode="before")
    @on_restore(handler=restore_default(10000))
    def validate_c_zero_pressure(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("b_weight", mode="before")
    @on_restore(handler=restore_default(3000))
    def validate_b_weight(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("b_diameter", mode="before")
    @on_restore(handler=restore_default(1800))
    def validate_b_diameter(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("b_length", mode="before")
    @on_restore(handler=restore_default(338))
    def validate_b_length(cls, value, info: FieldValidationInfo):
        return trigger_confield_validation(cls, value, info)

    @field_validator("twist_dir", mode="before")
    @on_restore(handler=restore_default(TwistDir.RIGHT.value))
    def validate_twist_dir(cls, value, info: FieldValidationInfo):
        if value not in [TwistDir.RIGHT, TwistDir.LEFT]:
            raise ValueError("Input should be 'RIGHT' or 'LEFT'")
        return value

    @field_validator("distances", mode="before")
    @on_restore(handler=restore_default(DEFAULT_DISTANCES))
    def validate_distances(cls, value, info: FieldValidationInfo):
        if not isinstance(value, list):
            raise ValueError("Input should be a valid list")

        if len(value) < 1:
            raise ValueError(
                "List should have at least 1 item after validation, not %s" % len(value)
            )
        if len(value) > 200:
            raise ValueError(
                "List should have at most 200 items after validation, not  %s"
                % len(value)
            )

        if len(value) != len(set(value)):
            repeated_items = list(
                set([item for item in value if value.count(item) > 1])
            )

            raise ValueError("Non unique values found in a list %s" % repeated_items)

        validator = pre_validate_conint(*get_args(Distance))

        for d in value:
            try:
                validator(d)
            except (TypeError, ValueError) as err:
                raise ValueError("Invalid values found: %s" % err)

        return value

    @field_validator("c_zero_distance_idx", mode="before")
    @on_restore(handler=restore_default(0))
    def validate_c_zero_distance_idx(cls, value, info: FieldValidationInfo):
        # Retrieve 'distances' from the model's data
        distances = info.data.get("distances")

        # If distances is None, raise an error since validation cannot proceed
        if distances is None:
            raise ValueError(
                "c_zero_distance_idx cannot be valid if distances are invalid or missing"
            )

        # If distances is a list or tuple, and value is not None, validate the index
        if isinstance(distances, (list, tuple)) and value is not None:
            if value >= len(distances):  # Check if index is out of bounds
                raise ValueError(
                    f"c_zero_distance_idx must be less than the length of distances (length={len(distances)})"
                )

        return value

    @field_validator("switches", mode="before")
    @on_restore(handler=restore_default(DEFAULT_SWITCHES))
    def validate_switches(cls, value, info: FieldValidationInfo):
        value.append(value[0])
        value[0]["c_idx"] = 500
        if not isinstance(value, list):
            raise ValueError("Input should be a valid list")

        if len(value) < 4:
            raise ValueError(
                "List should have at least 4 item after validation, not %s" % len(value)
            )

        for item in value:
            try:
                Switch.model_validate(item, context=info.context)
            except ValidationError as err:
                raise ValueError("Invalid values found: %s" % err)

        return value

    @field_validator("bc_type", mode="before")
    @on_restore(handler=restore_default(BCType.G7))
    def validate_bc_type(cls, value, info: FieldValidationInfo):
        if value not in [BCType.G1, BCType.G7, BCType.CUSTOM]:
            raise ValueError("Input should be 'G1', 'G7' or 'CUSTOM'")
        return value

    @field_validator("coef_rows", mode="before")
    @on_restore(handler=restore_default(DEFAULT_COEF_ROWS))
    def validate_coef_rows_based_on_bc_type(cls, value, info: FieldValidationInfo):
        # Convert dictionaries to model instances based on bc_type
        bc_type = info.data.get("bc_type")
        value[0]["bc_cd"] = 300000
        try:
            if bc_type in [BCType.G1, BCType.G7]:
                if len(value) < 1:
                    raise ValueError(
                        "coef_rows should have at least 1 item when bc_type is %s"
                        % bc_type
                    )
                if len(value) > 5:
                    raise ValueError(
                        "coef_rows should have maximum 5 items when bc_type is %s"
                        % bc_type
                    )

                # Ensure coef_rows contains BcMvRows instances
                value = [
                    BcMvRows(**row) if isinstance(row, dict) else row for row in value
                ]
                if not all(isinstance(row, BcMvRows) for row in value):
                    raise ValueError(
                        "coef_rows must contain BcMvRows when bc_type is %s" % bc_type
                    )
            elif bc_type == BCType.CUSTOM:
                if len(value) < 1:
                    raise ValueError(
                        "coef_rows should have at least 1 item when bc_type is {bc_type}"
                    )
                if len(value) > 200:
                    raise ValueError(
                        "coef_rows should have maximum 200 items when bc_type is {bc_type}"
                    )

                # Ensure coef_rows contains CdMaRows instances
                value = [
                    CdMaRows(**row) if isinstance(row, dict) else row for row in value
                ]
                if not all(isinstance(row, CdMaRows) for row in value):
                    raise ValueError(
                        "coef_rows must contain CdMaRows when bc_type is %s" % bc_type
                    )
            else:
                raise ValueError("Unsupported bc_type: %s" % bc_type)

        except ValidationError:
            # raise e
            # Handle the validation error gracefully
            raise ValueError(
                "Validation error while processing coef_rows for bc_type %s" % bc_type
            )

        return value


class Payload(BaseModel):
    profile: Profile



# Copyright 2023 Buf Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import celpy  # type: ignore
from celpy import celtypes  # type: ignore

QUOTE_TRANS = str.maketrans(
    {
        "\a": r"\a",
        "\b": r"\b",
        "\f": r"\f",
        "\n": r"\n",
        "\r": r"\r",
        "\t": r"\t",
        "\v": r"\v",
        "\\": r"\\",
        '"': r"\"",
    }
)


def quote(s: str) -> str:
    return '"' + s.translate(QUOTE_TRANS) + '"'


class StringFormat:
    """An implementation of string.format() in CEL."""

    def __init__(self, locale: str):
        self.locale = locale

    def format(self, fmt: celtypes.Value, args: celtypes.Value) -> celpy.Result:  # noqa: A003
        if not isinstance(fmt, celtypes.StringType):
            return celpy.native_to_cel(celpy.new_error("format() requires a string as the first argument"))
        if not isinstance(args, celtypes.ListType):
            return celpy.native_to_cel(celpy.new_error("format() requires a list as the second argument"))
        # printf style formatting
        i = 0
        j = 0
        result = ""
        while i < len(fmt):
            if fmt[i] != "%":
                result += fmt[i]
                i += 1
                continue

            if i + 1 < len(fmt) and fmt[i + 1] == "%":
                result += "%"
                i += 2
                continue
            if j >= len(args):
                return celpy.CELEvalError("format() not enough arguments for format string")
            arg = args[j]
            j += 1
            i += 1
            if i >= len(fmt):
                return celpy.CELEvalError("format() incomplete format specifier")
            precision = 6
            if fmt[i] == ".":
                i += 1
                precision = 0
                while i < len(fmt) and fmt[i].isdigit():
                    precision = precision * 10 + int(fmt[i])
                    i += 1
            if i >= len(fmt):
                return celpy.CELEvalError("format() incomplete format specifier")
            if fmt[i] == "f":
                result += self.format_float(arg, precision)
            if fmt[i] == "e":
                result += self.format_exponential(arg, precision)
            elif fmt[i] == "d":
                result += self.format_int(arg)
            elif fmt[i] == "s":
                result += self.format_string(arg)
            elif fmt[i] == "x":
                result += self.format_hex(arg)
            elif fmt[i] == "X":
                result += self.format_hex(arg).upper()
            elif fmt[i] == "o":
                result += self.format_oct(arg)
            elif fmt[i] == "b":
                result += self.format_bin(arg)
            else:
                return celpy.CELEvalError("format() unknown format specifier: " + fmt[i])
            i += 1
        if j < len(args):
            return celpy.CELEvalError("format() too many arguments for format string")
        return celtypes.StringType(result)

    def format_float(self, arg: celtypes.Value, precision: int) -> celpy.Result:
        if isinstance(arg, celtypes.DoubleType):
            return celtypes.StringType(f"{arg:.{precision}f}")
        return self.format_int(arg)

    def format_exponential(self, arg: celtypes.Value, precision: int) -> celpy.Result:
        if isinstance(arg, celtypes.DoubleType):
            return celtypes.StringType(f"{arg:.{precision}e}")
        return self.format_int(arg)

    def format_int(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, celtypes.IntType):
            return celtypes.StringType(arg)
        if isinstance(arg, celtypes.UintType):
            return celtypes.StringType(arg)
        return celpy.CELEvalError("format_int() requires an integer argument")

    def format_hex(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, celtypes.IntType):
            return celtypes.StringType(f"{arg:x}")
        if isinstance(arg, celtypes.UintType):
            return celtypes.StringType(f"{arg:x}")
        if isinstance(arg, celtypes.BytesType):
            return celtypes.StringType(arg.hex())
        if isinstance(arg, celtypes.StringType):
            return celtypes.StringType(arg.encode("utf-8").hex())
        return celpy.CELEvalError("format_hex() requires an integer, string, or binary argument")

    def format_oct(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, celtypes.IntType):
            return celtypes.StringType(f"{arg:o}")
        if isinstance(arg, celtypes.UintType):
            return celtypes.StringType(f"{arg:o}")
        return celpy.CELEvalError("format_oct() requires an integer argument")

    def format_bin(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, celtypes.IntType):
            return celtypes.StringType(f"{arg:b}")
        if isinstance(arg, celtypes.UintType):
            return celtypes.StringType(f"{arg:b}")
        if isinstance(arg, celtypes.BoolType):
            return celtypes.StringType(f"{arg:b}")
        return celpy.CELEvalError("format_bin() requires an integer argument")

    def format_string(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, celtypes.StringType):
            return arg
        if isinstance(arg, celtypes.BytesType):
            return celtypes.StringType(arg.hex())
        if isinstance(arg, celtypes.ListType):
            return self.format_list(arg)
        return celtypes.StringType(arg)

    def format_value(self, arg: celtypes.Value) -> celpy.Result:
        if isinstance(arg, (celtypes.StringType, str)):
            return celtypes.StringType(quote(arg))
        if isinstance(arg, celtypes.UintType):
            return celtypes.StringType(arg)
        return self.format_string(arg)

    def format_list(self, arg: celtypes.ListType) -> celpy.Result:
        result = "["
        for i in range(len(arg)):
            if i > 0:
                result += ", "
            result += self.format_value(arg[i])
        result += "]"
        return celtypes.StringType(result)


_default_format = StringFormat("en_US")
format = _default_format.format  # noqa: A001
format_value = _default_format.format_value



import sys
from argparse import ArgumentParser
from dataclasses import dataclass
from importlib import metadata
from pathlib import Path

import tqdm

import a7p
from a7p import exceptions, profedit_pb2, setUseProtovalidate
from a7p.a7p import setUseSpecValidator
from a7p.exceptions import A7PValidationError
from a7p.factory import DistanceTable
from a7p.logger import logger, color_print, color_fmt
from a7p.recover.recover_process import attempt_to_recover

try:
    __version__ = metadata.version("a7p")
except metadata.PackageNotFoundError:
    __version__ = "undefined version"

DISTANCES = {
    "subsonic": DistanceTable.SUBSONIC.value,
    "low": DistanceTable.LOW_RANGE.value,
    "medium": DistanceTable.MEDIUM_RANGE.value,
    "long": DistanceTable.LONG_RANGE.value,
    "ultra": DistanceTable.ULTRA_RANGE.value,
}


class CustomArgumentParser(ArgumentParser):
    def error(self, message):
        """Override error method to show help message on argument errors."""
        self.print_help(sys.stderr)
        # sys.stderr.write(f"\nError: {message}\n")
        logger.error(f"Error: {message}")
        sys.exit(2)

    def warning(self, message):
        """Override error method to show help message on argument errors."""
        self.print_help(sys.stderr)
        # sys.stderr.write(f"\nError: {message}\n")
        logger.warning(f"Error: {message}")
        sys.exit(2)


parser = CustomArgumentParser(
    f"a7p {__version__}",
    exit_on_error=True,
)

parser.add_argument(
    "path",
    type=Path,
    help="Specify the path to the directory or a .a7p file to process.",
)

# Main options
parser.add_argument(
    "-V",
    "--version",
    action="version",
    version=__version__,
    help="Display the current version of the tool.",
)
parser.add_argument(
    "-r",
    "--recursive",
    action="store_true",
    help="Recursively process files in the specified directory.",
)
parser.add_argument(
    "-F",
    "--force",
    action="store_true",
    help="Force saving changes without confirmation.",
)
parser.add_argument(
    "--unsafe", action="store_true", help="Skip data validation (use with caution)."
)

recover_group = parser.add_argument_group("Single file specific options")
recover_group.add_argument(
    "--verbose",
    action="store_true",
    help="Enable verbose output for detailed logs. "
    "This option is only allowed for a single file.",
)
recover_group.add_argument(
    "--recover",
    action="store_true",
    help="Attempt to recover from errors found in a file. "
    "This option is only allowed for a single file.",
)

# Distances group
distances_group = parser.add_argument_group("Distances")
distances_group.add_argument(
    "-zd",
    "--zero-distance",
    action="store",
    type=int,
    help="Set the zero distance in meters.",
)
distances_group.add_argument(
    "-d",
    "--distances",
    action="store",
    choices=["subsonic", "low", "medium", "long", "ultra"],
    help="Specify the distance range: 'subsonic', 'low', 'medium', 'long', or 'ultra'.",
)

# Zeroing group
zeroing_group = parser.add_argument_group("Zeroing")
zeroing_exclusive_group = zeroing_group.add_mutually_exclusive_group()
zeroing_exclusive_group.add_argument(
    "-zs",
    "--zero-sync",
    action="store",
    type=Path,
    help="Synchronize zero using a specified configuration file.",
)
zeroing_exclusive_group.add_argument(
    "-zo",
    "--zero-offset",
    action="store",
    nargs=2,
    type=float,
    metavar=("X_OFFSET", "Y_OFFSET"),
    help="Set the offset for zeroing in clicks (X_OFFSET and Y_OFFSET).",
)

# zeroing_exclusive_group.add_argument('-cs', '--clicks-switch', action='store', nargs=4, help="Switch clicks sizes",
#                                      metavar=("CUR_X", "NEW_X", "CUR_Y", "NEW_Y"))

# Switches group (Archer devices specific)
switches_group = parser.add_argument_group("Switches")
switches_exclusive_group = switches_group.add_mutually_exclusive_group()
switches_exclusive_group.add_argument(
    "-cps",
    "--copy-switches-from",
    action="store",
    type=Path,
    help="Copy switches from other a7p file.",
    dest="copy_switches",
)

advanced_group = parser.add_argument_group("Advanced")
advanced_group.add_argument("--spec-validator", action="store_true", help="Use spec-based validator (deprecated).")
advanced_group.add_argument("--protovalidate", action="store_true", help="Use protovalidate (deprecated).")


# parser.add_argument('--max-threads', action='store', type=int, default=5)


@dataclass
class Result:
    path: Path
    error = None
    validation_error: A7PValidationError | None = None
    zero: tuple[float, float] = None
    new_zero: tuple[float, float] = None
    zero_update: bool = False
    distances: str = None
    zero_distance: str = None
    recover: bool = False
    payload: profedit_pb2.Payload = None
    switches: bool = False

    def reset_errors(self):
        self.error = None
        self.validation_error = None

    def print(self, verbose=False):
        if self.error:
            error = color_fmt(f"Invalid ({self.error}):", levelname="ERROR")
            print(f"{error} File: {self.path.absolute()}")

        else:
            print(
                f"{color_fmt('Valid:', levelname='INFO')} File: {self.path.absolute()}"
            )

        if self.zero:
            x, y = self.zero
            print("\tZero:\tX: {},\tY: {}".format(-x, y))
        if self.zero_update:
            x, y = self.new_zero
            color_print(
                "\tNew zero:\tX: {},\tY: {}".format(-x, y), levelname="LIGHT_BLUE"
            )
        if self.distances:
            color_print(
                "\tNew range: {}".format(self.distances), levelname="LIGHT_BLUE"
            )
        if self.zero_distance:
            color_print(
                "\tNew zero distance: {}".format(self.zero_distance),
                levelname="LIGHT_BLUE",
            )
        if self.switches:
            color_print("\tSwitches copied", levelname="LIGHT_BLUE")

        if self.validation_error and verbose:
            for violation in self.validation_error.all_violations:
                color_print(violation.format(), levelname="WARNING")

    def save_changes(self, force=False):
        if (
            self.zero_distance
            or self.distances
            or self.zero_update
            or self.recover
            or self.switches
        ):
            if not force:
                yes_no = input(
                    color_fmt(
                        "Do you want to save changes? (Y/N): ", levelname="LIGHT_YELLOW"
                    )
                )
                if yes_no.lower() != "y":
                    logger.info("No changes have been saved.")
                    return
            try:
                try:
                    # Serialize and validate the payload
                    data = a7p.dumps(self.payload)

                    # Write the validated data to the file
                    filepath = self.path.absolute()
                    if self.recover:
                        filepath = filepath.with_name(
                            filepath.stem + "_recovered" + filepath.suffix
                        )

                    with open(filepath, "wb") as fp:
                        fp.write(data)
                        logger.info(
                            f"Changes have been saved successfully to {filepath}."
                        )
                except exceptions.A7PDataError:
                    logger.warning("The data is invalid. Changes have not been saved.")
            except IOError as e:
                logger.warning(f"An error occurred while saving: {e}")


def update_distances(payload, distances, zero_distance):
    if not zero_distance:
        cur_zero_distance = payload.profile.distances[
            payload.profile.c_zero_distance_idx
        ]
    else:
        cur_zero_distance = zero_distance * 100
    if distances:
        new_distances = [int(d * 100) for d in DISTANCES[distances]]
    else:
        new_distances = payload.profile.distances[:]
    new_distances.append(cur_zero_distance)
    new_distances = list(set(new_distances))
    new_distances.sort()
    payload.profile.distances[:] = new_distances
    payload.profile.c_zero_distance_idx = new_distances.index(cur_zero_distance)


def update_zeroing(payload, zero_offset=None, zero_sync=None):
    if zero_offset:
        x_offset, y_offset = zero_offset
        payload.profile.zero_x = payload.profile.zero_x + round(x_offset * -1000)
        payload.profile.zero_y = payload.profile.zero_y + round(y_offset * 1000)
    elif zero_sync:
        x_zero, y_zero = zero_sync
        payload.profile.zero_x = x_zero
        payload.profile.zero_y = y_zero


def update_switches(payload, switches):
    if switches:
        del payload.profile.switches[:]
        payload.profile.switches.extend(switches)


def update_data(
    payload, distances, zero_distance, zero_offset, zero_sync, copy_switches
):
    update_distances(payload, distances, zero_distance)
    update_zeroing(payload, zero_offset, zero_sync)
    update_switches(payload, copy_switches)


def get_zero_to_sync(path, validate):
    try:
        with open(path, "rb") as f:
            payload = a7p.load(f, validate_=validate, fail_fast=True)
        return payload.profile.zero_x, payload.profile.zero_y
    except (IOError, exceptions.A7PDataError) as e:
        parser.error(e)


def get_switches_to_copy(path, validate):
    try:
        with open(path, "rb") as f:
            payload = a7p.load(f, validate_=validate, fail_fast=True)
        return payload.profile.switches
    except (IOError, exceptions.A7PDataError) as e:
        parser.error(e)


def process_file(
    path,
    validate=False,
    distances=None,
    zero_distance=None,
    zero_offset=None,
    zero_sync=None,
    verbose=False,
    recover=False,
    copy_switches=None,
):
    if path.suffix != ".a7p":
        return
    result = Result(
        path,
        distances=distances,
        zero_distance=zero_distance,
        zero_update=any([zero_offset, zero_sync]),
        switches=copy_switches is not None,
    )
    try:
        with open(path, "rb") as fp:
            data = fp.read()
        try:
            payload = a7p.loads(data, validate_=validate, fail_fast=not verbose)
        except exceptions.A7PValidationError as err:
            result.error = "Validation error"
            result.validation_error = err
            payload = err.payload
    except (IOError, exceptions.A7PDataError) as err:
        result.error = err
        return result

    result.zero = (payload.profile.zero_x / 1000, payload.profile.zero_y / 1000)
    if distances or zero_distance or result.zero_update or copy_switches:
        update_data(
            payload, distances, zero_distance, zero_offset, zero_sync, copy_switches
        )

    result.new_zero = (payload.profile.zero_x / 1000, payload.profile.zero_y / 1000)

    result.payload = payload

    if recover:
        recover_payload(result)

    return result


def print_results_and_save(results, verbose=False, force=False):
    count_errors = 0
    results = sorted(
        filter(lambda x: x is not None, results),
        key=lambda x: x.error is not None,
        reverse=False,
    )
    for result in results:
        if result:
            result.print(verbose)
            print()
            result.save_changes(force)

            if result.error:
                count_errors += 1

    output_strings = [
        f"Files checked: {len(results)}",
        color_fmt(f"Ok: {len(results) - count_errors}", levelname="INFO"),
        color_fmt(f"Failed: {count_errors}", levelname="ERROR"),
    ]
    print(", ".join(output_strings))


def recover_payload(result: Result):
    if result.validation_error:
        result.recover = True

        color_print("Violations found:", levelname="ERROR")
        for v in result.validation_error.all_violations:
            color_print(v.format(), levelname="WARNING")

        recover_error = attempt_to_recover(result.validation_error)
        if not recover_error:
            result.reset_errors()

    else:
        logger.info("No violations found")


def process_files(
    path: Path = None,
    recursive: bool = False,
    unsafe: bool = False,
    distances: str = None,
    zero_distance: int = None,
    json: Path = None,
    verbose: bool = False,
    force: bool = False,
    zero_offset: tuple[float, float] = None,
    zero_sync: Path = None,
    recover: bool = False,
    copy_switches: Path = None,
):
    if not Path.exists(path):
        parser.warning(f"The '{path}' is not a valid path")
    if unsafe:
        logger.warning(
            "The 'unsafe' mode is restricted and may lead to file corruption."
        )
    if force:
        logger.warning(
            "Use the 'force' option cautiously, only if you are certain about its effects."
        )

    validate = unsafe is False

    if zero_sync:
        zero_sync = get_zero_to_sync(zero_sync, validate)

    if copy_switches:
        copy_switches = get_switches_to_copy(copy_switches, validate)

    if not path.is_dir():
        if path and recover:
            invalid_options_for_recover = [
                recursive,
                unsafe,
                distances,
                zero_distance,
                json,
                verbose,
                force,
                zero_offset,
                zero_sync,
                copy_switches,
            ]

            if any(invalid_options_for_recover):
                raise parser.error(
                    "The '--recover' option cannot be combined with other options."
                )

            verbose = True

        results = [
            process_file(
                path,
                validate,
                distances,
                zero_distance,
                zero_offset,
                zero_sync,
                verbose,
                recover,
                copy_switches,
            )
        ]
    else:
        if recover:
            parser.warning(
                "The '--recover' option is supported only when processing a single file."
            )
        if verbose:
            parser.warning(
                "The '--verbose' option is supported only when processing a single file."
            )

        results: tuple[Result] | list[Result] = []

        if recursive:
            paths = path.rglob("*")
        else:
            paths = path.iterdir()

        files = tuple(p for p in paths if p.is_file())
        for item in tqdm.tqdm(files):
            results.append(
                process_file(
                    item,
                    validate,
                    distances,
                    zero_distance,
                    zero_offset,
                    zero_sync,
                    copy_switches=copy_switches,
                )
            )

    print_results_and_save(results, verbose=verbose, force=force)


def main():
    try:
        args = parser.parse_args()

        args_dict = args.__dict__
        use_proto = args_dict.pop("protovalidate", False)
        use_spec = args_dict.pop("spec_validator", False)
        setUseProtovalidate(use_proto)
        setUseSpecValidator(use_spec)

        process_files(**args.__dict__)
    except NotImplementedError as e:
        logger.error(e)
        sys.exit(0)
    except Exception as e:
        logger.exception(e)
        logger.critical(e)
        sys.exit(1)
    except KeyboardInterrupt:
        print()
        logger.warning("Process interrupted by user. Exiting gracefully...")
        sys.exit(0)
    logger.info("Process completed successfully. Exiting gracefully...")
    sys.exit(0)


if __name__ == "__main__":
    main()



from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

from google._upb._message import RepeatedScalarContainer, RepeatedCompositeContainer
from typing_extensions import Any

from a7p import profedit_pb2, A7PFactory
from a7p.factory import Switches
from a7p.logger import color_fmt, logger


@dataclass
class RecoverResult:
    recovered: bool
    path: Path | str
    old_value: str | None = None
    new_value: Any = None

    def print(self):
        if isinstance(self.path, Path):
            path_string = f"{self.path.as_posix()}"
        else:
            path_string = f"{self.path}"

        if len(path_string) > 30:
            path_string = path_string[:27] + "..."
        path_string = path_string.ljust(30)

        if self.recovered:
            prefix = color_fmt("Recovered".ljust(10), levelname="INFO")
        else:
            prefix = color_fmt("Skipped".ljust(10), levelname="WARNING")

        def truncate_list(_value: Any) -> str:
            if isinstance(
                _value,
                (RepeatedScalarContainer, RepeatedCompositeContainer, list, tuple),
            ):
                _value = [str(v) for v in _value]
                if len(_value) > 6:
                    _value = (
                        f"[ {', '.join(_value[:3])}, ... {', '.join(_value[-3:])} ]"
                    )
                else:
                    _value = f"[ {','.join(_value)} ]"
            return _value

        def truncate(string: str):
            string = str(string).replace("\n", " ")
            if len(string) > 50:
                # string = f'[ {string[:25]} ... {string[-25:]} ]'
                string = f"{string[:25]} ... {string[-25:]}"
            return string

        old_value = str(truncate_list(self.old_value))
        new_value = str(truncate_list(self.new_value))

        print(
            f"{prefix} : {path_string} : value : {truncate(old_value)} -> {truncate(new_value)}"
        )


class Recover:
    def __init__(self):
        self.recover_funcs = {}

    def register(self, path, func):
        raise NotImplementedError("register not implemented as is abstract method")

    @staticmethod
    def split_path(path: Path | str) -> list:
        raise NotImplementedError("split_path not implemented as is abstract method")

    @classmethod
    def get_value_by_violation(cls, payload, violation):
        # raise NotImplementedError("get_value_by_violation not implemented as is abstract method")
        _value = payload
        # _path = violation.path.split('.')
        _path = cls.split_path(violation.path)
        for p in _path:
            if hasattr(_value, p):
                _value = getattr(_value, p)

        return deepcopy(_value)

    def recover_one(self, payload, violation):
        if violation.path in self.recover_funcs:
            old_value = self.get_value_by_violation(payload, violation)
            self.recover_funcs[violation.path](payload)
            new_value = self.get_value_by_violation(payload, violation)
            return RecoverResult(True, violation.path, old_value, new_value)

        return RecoverResult(False, violation.path, None, None)

    def recover(self, payload, violations):
        results = []
        for v in violations:
            result = self.recover_one(payload, v)
            results.append(result)
        return results


def fix_str_len_type(string: str, expected_len: int, default: str = "nil"):
    if isinstance(string, str):
        return string[:expected_len]
    return default[:expected_len]


def _recover_proto_bullet_name(payload):
    payload.profile.bullet_name = fix_str_len_type(payload.profile.bullet_name, 50)


def _recover_proto_cartridge_name(payload):
    payload.profile.cartridge_name = fix_str_len_type(
        payload.profile.cartridge_name, 50
    )


def _recover_proto_profile_name(payload):
    payload.profile.profile_name = fix_str_len_type(payload.profile.profile_name, 50)


def _recover_proto_user_note(payload):
    payload.profile.user_note = fix_str_len_type(
        payload.profile.user_note, 250, "Warning: Restored profile"
    )


def _recover_proto_short_name_top(payload):
    payload.profile.short_name_top = fix_str_len_type(payload.profile.short_name_top, 8)


def _recover_proto_short_name_bot(payload):
    payload.profile.short_name_bot = fix_str_len_type(payload.profile.short_name_bot, 8)


def _recover_proto_zero_x(payload):
    payload.profile.zero_x = 0


def _recover_proto_zero_y(payload):
    payload.profile.zero_y = 0


def _recover_proto_sc_height(payload):
    payload.profile.sc_height = 90


def _recover_proto_r_twist(payload):
    payload.profile.r_twist = 10


def _recover_proto_c_muzzle_velocity(payload):
    payload.profile.c_muzzle_velocity = 8000


def _recover_proto_c_zero_temperature(payload):
    payload.profile.c_zero_temperature = 15


def _recover_proto_c_t_coeff(payload):
    payload.profile.c_t_coeff = 1000


def _recover_proto_c_zero_distance_idx(payload):
    payload.profile.c_zero_distance_idx = 0


def _recover_proto_c_zero_air_temperature(payload):
    payload.profile.c_zero_air_temperature = 15


def _recover_proto_c_zero_air_pressure(payload):
    payload.profile.c_zero_air_pressure = 10000


def _recover_proto_c_zero_air_humidity(payload):
    payload.profile.c_zero_air_pressure = 0


def _recover_proto_c_zero_w_pitch(payload):
    payload.profile.c_zero_w_pitch = 0


def _recover_proto_c_zero_p_temperature(payload):
    payload.profile.c_zero_p_temperature = 15


def _recover_proto_b_diameter(payload):
    payload.profile.b_diameter = 338


def _recover_proto_b_weight(payload):
    payload.profile.b_weight = 3000


def _recover_proto_b_length(payload):
    payload.profile.b_length = 1700


def _recover_proto_twist_dir(payload):
    payload.twist_dir = profedit_pb2.TwistDir.RIGHT


def _recover_proto_bc_type(payload):
    logger.warning("Drag model restored to G7")
    payload.bc_type = profedit_pb2.GType.G7


def _recover_proto_switches(payload):
    del payload.profile.switches[:]
    payload.profile.switches.extend(Switches())


def _recover_proto_coef_rows(payload):
    logger.warning("Drag model coefficients restored to 0.1")
    del payload.profile.coef_rows[:]
    payload.profile.coef_rows.extend(
        profedit_pb2.CoefRow(bc_cd=round(0.1 * 10000), mv=round(0 * 10))
    )


def _recover_proto_distances(payload):
    del payload.profile.distances[:]
    payload.profile.distances[:] = [
        int(d * 100) for d in A7PFactory.DistanceTable.LONG_RANGE.value
    ]


def _recover_proto_caliber(payload):
    payload.profile.caliber = fix_str_len_type(payload.profile.caliber, 50)


def _recover_proto_device_uuid(payload):
    payload.profile.device_uuid = fix_str_len_type(payload.profile.device_uuid, 50, "")


class RecoverProto(Recover):
    def register(self, path, func):
        self.recover_funcs[path] = func

    @staticmethod
    def split_path(path: Path | str) -> list:
        return path.split(".")


recover_proto = RecoverProto()

recover_proto.register("profile.profile_name", _recover_proto_short_name_top)
recover_proto.register("profile.cartridge_name", _recover_proto_cartridge_name)
recover_proto.register("profile.bullet_name", _recover_proto_bullet_name)
recover_proto.register("profile.user_note", _recover_proto_user_note)

recover_proto.register("profile.short_name_top", _recover_proto_short_name_top)
recover_proto.register("profile.short_name_bot", _recover_proto_short_name_bot)

recover_proto.register("profile.zero_x", _recover_proto_zero_x)
recover_proto.register("profile.zero_y", _recover_proto_zero_y)

recover_proto.register("profile.sc_height", _recover_proto_sc_height)
recover_proto.register("profile.r_twist", _recover_proto_r_twist)

recover_proto.register("profile.c_muzzle_velocity", _recover_proto_c_muzzle_velocity)
recover_proto.register("profile.c_zero_temperature", _recover_proto_c_zero_temperature)
recover_proto.register("profile.c_t_coeff", _recover_proto_c_t_coeff)
recover_proto.register(
    "profile.c_zero_distance_idx", _recover_proto_c_zero_distance_idx
)
recover_proto.register(
    "profile.c_zero_air_temperature", _recover_proto_c_zero_air_temperature
)
recover_proto.register(
    "profile.c_zero_air_pressure", _recover_proto_c_zero_air_pressure
)
recover_proto.register(
    "profile.c_zero_air_humidity", _recover_proto_c_zero_air_humidity
)
recover_proto.register("profile.c_zero_w_pitch", _recover_proto_c_zero_w_pitch)
recover_proto.register(
    "profile.c_zero_p_temperature", _recover_proto_c_zero_p_temperature
)

recover_proto.register("profile.b_diameter", _recover_proto_b_diameter)
recover_proto.register("profile.b_weight", _recover_proto_b_weight)
recover_proto.register("profile.b_length", _recover_proto_b_length)

recover_proto.register("profile.twist_dir", _recover_proto_twist_dir)
recover_proto.register("profile.bc_type", _recover_proto_bc_type)

recover_proto.register("profile.switches", _recover_proto_switches)
recover_proto.register("profile.distances", _recover_proto_distances)
recover_proto.register("profile.coef_rows", _recover_proto_coef_rows)
recover_proto.register("profile.caliber", _recover_proto_caliber)
recover_proto.register("profile.device_uuid", _recover_proto_device_uuid)


class RecoverSpec(Recover):
    def register(self, path, func):
        self.recover_funcs[Path(path)] = func

    @staticmethod
    def split_path(path: Path | str) -> list:
        return Path(path).as_posix().split("/")[1:]


def _recover_spec_bullet_name(payload):
    payload.profile.bullet_name = fix_str_len_type(payload.profile.bullet_name, 49)


def _recover_spec_cartridge_name(payload):
    payload.profile.cartridge_name = fix_str_len_type(
        payload.profile.cartridge_name, 49
    )


def _recover_spec_caliber(payload):
    payload.profile.caliber = fix_str_len_type(payload.profile.caliber, 49)


def _recover_spec_profile_name(payload):
    payload.profile.profile_name = fix_str_len_type(payload.profile.profile_name, 49)


def _recover_spec_uuid(payload):
    payload.profile.device_uuid = fix_str_len_type(payload.profile.device_uuid, 49)


def _recover_spec_user_note(payload):
    payload.profile.user_note = fix_str_len_type(
        payload.profile.user_note, 1023, "Warning: Restored profile"
    )


def _recover_spec_short_name_top(payload):
    payload.profile.short_name_top = fix_str_len_type(payload.profile.short_name_top, 7)


def _recover_spec_short_name_bot(payload):
    payload.profile.short_name_bot = fix_str_len_type(payload.profile.short_name_bot, 7)


def _recover_spec_zero_x(payload):
    payload.profile.zero_x = 0


def _recover_spec_zero_y(payload):
    payload.profile.zero_y = 0


def _recover_spec_sc_height(payload):
    payload.profile.sc_height = 90


def _recover_spec_r_twist(payload):
    payload.profile.r_twist = 10


def _recover_spec_c_muzzle_velocity(payload):
    payload.profile.c_muzzle_velocity = 8000


def _recover_spec_c_zero_temperature(payload):
    payload.profile.c_zero_temperature = 15


def _recover_spec_c_t_coeff(payload):
    payload.profile.c_t_coeff = 1000


def _recover_spec_c_zero_distance_idx(payload):
    payload.profile.c_zero_distance_idx = 0


def _recover_spec_c_zero_air_temperature(payload):
    payload.profile.c_zero_air_temperature = 15


def _recover_spec_c_zero_air_pressure(payload):
    payload.profile.c_zero_air_pressure = 10000


def _recover_spec_c_zero_air_humidity(payload):
    payload.profile.c_zero_air_pressure = 0


def _recover_spec_c_zero_w_pitch(payload):
    payload.profile.c_zero_w_pitch = 0


def _recover_spec_c_zero_p_temperature(payload):
    payload.profile.c_zero_p_temperature = 15


def _recover_spec_b_diameter(payload):
    payload.profile.b_diameter = 338


def _recover_spec_b_weight(payload):
    payload.profile.b_weight = 3000


def _recover_spec_b_length(payload):
    payload.profile.b_length = 1700


def _recover_spec_twist_dir(payload):
    payload.twist_dir = profedit_pb2.TwistDir.RIGHT


def _recover_spec_bc_type(payload):
    logger.warning("Drag model restored to G7")
    payload.bc_type = profedit_pb2.GType.G7


def _recover_spec_switches(payload):
    del payload.profile.switches[:]
    payload.profile.switches.extend(Switches())


def _recover_spec_coef_rows(payload):
    logger.warning("Drag model coefficients restored to 0.1")
    del payload.profile.coef_rows[:]
    payload.profile.coef_rows.extend(
        profedit_pb2.CoefRow(bc_cd=round(0.1 * 10000), mv=round(0 * 10))
    )


def _recover_spec_distances(payload):
    del payload.profile.distances[:]
    payload.profile.distances[:] = [
        int(d * 100) for d in A7PFactory.DistanceTable.LONG_RANGE.value
    ]


recover_spec = RecoverSpec()

recover_spec.register("~/profile/profile_name", _recover_spec_profile_name)
recover_spec.register("~/profile/cartridge_name", _recover_spec_cartridge_name)
recover_spec.register("~/profile/bullet_name", _recover_spec_bullet_name)
recover_spec.register("~/profile/user_note", _recover_spec_user_note)
recover_spec.register("~/profile/device_uuid", _recover_spec_uuid)
recover_spec.register("~/profile/short_name_top", _recover_spec_short_name_top)
recover_spec.register("~/profile/short_name_bot", _recover_spec_short_name_bot)

recover_spec.register("~/profile/zero_x", _recover_spec_zero_x)
recover_spec.register("~/profile/zero_y", _recover_spec_zero_y)

recover_spec.register("~/profile/sc_height", _recover_spec_sc_height)
recover_spec.register("~/profile/r_twist", _recover_spec_r_twist)

recover_spec.register("~/profile/c_muzzle_velocity", _recover_spec_c_muzzle_velocity)
recover_spec.register("~/profile/c_zero_temperature", _recover_spec_c_zero_temperature)
recover_spec.register("~/profile/c_t_coeff", _recover_spec_c_t_coeff)
recover_spec.register(
    "~/profile/c_zero_distance_idx", _recover_spec_c_zero_distance_idx
)
recover_spec.register(
    "~/profile/c_zero_air_temperature", _recover_spec_c_zero_air_temperature
)
recover_spec.register(
    "~/profile/c_zero_air_pressure", _recover_spec_c_zero_air_pressure
)
recover_spec.register(
    "~/profile/c_zero_air_humidity", _recover_spec_c_zero_air_humidity
)
recover_spec.register("~/profile/c_zero_w_pitch", _recover_spec_c_zero_w_pitch)
recover_spec.register(
    "~/profile/c_zero_p_temperature", _recover_spec_c_zero_p_temperature
)

recover_spec.register("~/profile/b_diameter", _recover_spec_b_diameter)
recover_spec.register("~/profile/b_weight", _recover_spec_b_weight)
recover_spec.register("~/profile/b_length", _recover_spec_b_length)

recover_spec.register("~/profile/twist_dir", _recover_spec_twist_dir)
recover_spec.register("~/profile/bc_type", _recover_spec_bc_type)

recover_spec.register("~/profile/distances", _recover_spec_distances)
recover_spec.register("~/profile/switches", _recover_spec_switches)
recover_spec.register("~/profile/coef_rows", _recover_spec_coef_rows)



from typing import Any

from pydantic import ValidationError
from typing_extensions import List, Dict

import a7p
from a7p import exceptions, profedit_pb2
from a7p.pydantic.models import Payload
from a7p.pydantic.template import PAYLOAD_RECOVERY_SCHEMA


def get_dict_field(payload_dict: Dict[str, Any], field_path: str):
    loc = field_path.split(".")
    current = payload_dict

    for part in loc:
        if part.isdigit():  # If part is a number (array index)
            current = current[int(part)]  # Access list by index
        else:
            current = current.get(
                part
            )  # Access dict by key (assuming proto_obj is a dict)

        if current is None:  # If the field doesn't exist
            raise KeyError("Field not found: %s" % field_path)

    return current  # Return the final value


def set_dict_field(payload_dict: Dict[str, Any], field_path: str, value: Any):
    loc = field_path.split(".")  # Split path into components
    current = payload_dict

    # Traverse the path until the second-to-last part
    for part in loc[:-1]:
        if part.isdigit():  # If part is a number (array index)
            current = current[int(part)]  # Access list element by index
        else:
            current = current.get(part)  # Access field using attribute name

        if current is None:
            raise KeyError("Field not found: %s" % field_path)

    # Now we're at the last part, set the value
    last_part = loc[-1]
    if last_part.isdigit():  # If the last part is a number (array index)
        current[int(last_part)] = value  # Set the list element at that index
    else:
        current[last_part] = value


def validate(payload: profedit_pb2.Payload, restore=False):
    payload_dict = a7p.to_dict(payload)
    context = {"restore": restore, "restored": []}
    model = None
    violations = []

    try:
        model = Payload.model_validate(payload_dict, context=context)
    except ValidationError as err:
        for error in err.errors():
            field_path = ".".join(map(str, error.get("loc", tuple())))
            violations.append(
                exceptions.Violation(
                    path=field_path,
                    value=error.get("input", None),
                    reason=error.get("msg", "Undefined error"),
                )
            )
        # raise exceptions.A7PValidationError(
        #     "Pydantic validation error",
        #     payload=a7p.from_dict(payload_dict),
        #     violations=violations
        # )
        return model, context.get("restored"), violations
    return model, context.get("restored"), violations


def recursive_recover(path: str, old_value: Any) -> Any:
    loc = path.split(".") if path else []
    recover_value = get_dict_field(PAYLOAD_RECOVERY_SCHEMA, path)

    if callable(recover_value):
        return recover_value(old_value)
    elif isinstance(recover_value, dict):
        return {k: recursive_recover(".".join(loc + [k]), None) for k in recover_value}
    elif isinstance(recover_value, list):
        return [
            recursive_recover(".".join(loc + [str(i)]), None)
            for i in range(len(recover_value))
        ]
    return recover_value


def recover(payload_dict: Dict[str, Any], violations: List[exceptions.Violation]):
    try:
        Payload.model_validate(
            payload_dict, context={"recovery": PAYLOAD_RECOVERY_SCHEMA}
        )
    except ValidationError as err:
        violations = []
        for error in err.errors():
            field_path = ".".join(map(str, error.get("loc", tuple())))
            violations.append(
                exceptions.Violation(
                    path=field_path,
                    value=error.get("input", None),
                    reason=error.get("msg", "Undefined error"),
                )
            )
        raise exceptions.A7PValidationError(
            "Pydantic validation error",
            payload=a7p.from_dict(payload_dict),
            violations=violations,
        )

    # results: List[RecoverResult] = []
    #
    # for violation in violations:
    #     path = violation.path
    #     try:
    #         old_value = get_dict_field(payload_dict, path)
    #
    #         try:
    #             new_value = recursive_recover(path, old_value)
    #             set_dict_field(payload_dict, path, new_value)
    #             results.append(RecoverResult(
    #                 recovered=True,
    #                 path=path,
    #                 old_value=old_value,
    #                 new_value=new_value,
    #             ))
    #         except (KeyError, ValueError, AttributeError) as e:
    #             # Log the error if needed
    #             results.append(RecoverResult(
    #                 recovered=False,
    #                 path=path,
    #                 old_value=old_value,
    #                 new_value=None,
    #             ))
    #     except KeyError as e:
    #         # Log missing keys
    #         results.append(RecoverResult(
    #             recovered=False,
    #             path=path,
    #             old_value=None,
    #             new_value=None,
    #         ))
    #
    # return results
